/*
 * ** header v3.0
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2018 Research Group Biomedical Physics,
 * Max-Planck-Institute for Dynamics and Self-Organization Göttingen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */
package caosdb.server.permissions;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map.Entry;

class EntityACLFactory extends AbstractEntityACLFactory<EntityACL> {

  @Override
  protected EntityACL create(final Collection<EntityACI> acis) {
    return new EntityACL(acis);
  }
}

public abstract class AbstractEntityACLFactory<T extends EntityACL> {

  private final HashMap<ResponsibleAgent, Long> normal_grants =
      new HashMap<ResponsibleAgent, Long>();
  private final HashMap<ResponsibleAgent, Long> priority_grants =
      new HashMap<ResponsibleAgent, Long>();
  private final HashMap<ResponsibleAgent, Long> normal_denials =
      new HashMap<ResponsibleAgent, Long>();
  private final HashMap<ResponsibleAgent, Long> priority_denials =
      new HashMap<ResponsibleAgent, Long>();

  public void grant(final ResponsibleAgent role, final int... permissionBitNumber) {
    grant(role, false, permissionBitNumber);
  }

  public void deny(final ResponsibleAgent role, final int... permissionBitNumber) {
    deny(role, false, permissionBitNumber);
  }

  public void grant(
      final ResponsibleAgent role, final boolean priority, final int... permissionBitNumber) {
    for (final int i : permissionBitNumber) {
      grant(role, priority, EntityPermission.getEntityPermission(i));
    }
  }

  public void deny(
      final ResponsibleAgent role, final boolean priority, final int... permissionBitNumber) {
    for (final int i : permissionBitNumber) {
      deny(role, priority, EntityPermission.getEntityPermission(i));
    }
  }

  public void grant(final ResponsibleAgent role, final String... permission) {
    grant(role, false, permission);
  }

  public void deny(final ResponsibleAgent role, final String... permission) {
    deny(role, false, permission);
  }

  public void grant(
      final ResponsibleAgent role, final boolean priority, final String... permission) {
    for (final String s : permission) {
      if (s.contains("*")) {
        grant(role, priority, EntityPermission.getPermissionsPerWildCard(s));
      } else {
        grant(role, priority, EntityPermission.getEntityPermission(s));
      }
    }
  }

  public void deny(
      final ResponsibleAgent role, final boolean priority, final String... permission) {
    for (final String s : permission) {
      if (s.contains("*")) {
        deny(role, priority, EntityPermission.getPermissionsPerWildCard(s));
      } else {
        deny(role, priority, EntityPermission.getEntityPermission(s));
      }
    }
  }

  public void deny(
      final ResponsibleAgent role,
      final boolean priority,
      final Collection<EntityPermission> permission) {
    for (final EntityPermission p : permission) {
      deny(role, priority, p);
    }
  }

  public void grant(
      final ResponsibleAgent role,
      final boolean priority,
      final Collection<EntityPermission> permission) {
    for (final EntityPermission p : permission) {
      grant(role, priority, p);
    }
  }

  public void grant(
      final ResponsibleAgent role, final boolean priority, final EntityPermission... permission) {
    if (priority) {
      addACI(this.priority_grants, role, permission);
    } else {
      addACI(this.normal_grants, role, permission);
    }
  }

  public void deny(
      final ResponsibleAgent role, final boolean priority, final EntityPermission... permission) {
    if (priority) {
      addACI(this.priority_denials, role, permission);
    } else {
      addACI(this.normal_denials, role, permission);
    }
  }

  private static void addACI(
      final HashMap<ResponsibleAgent, Long> map,
      final ResponsibleAgent role,
      final EntityPermission permission) {
    long bitSet = permission.getBitSet();
    try {
      bitSet |= map.get(role);
    } catch (final NullPointerException e) {
      // do nothing
    }
    map.put(role, bitSet);
  }

  private static void addACI(
      final HashMap<ResponsibleAgent, Long> map,
      final ResponsibleAgent role,
      final EntityPermission[] permission) {
    for (final EntityPermission p : permission) {
      addACI(map, role, p);
    }
  }

  private EntityACI[] toEntityACIArray(
      final HashMap<ResponsibleAgent, Long> map, final long modBitSet) {
    final EntityACI[] ret = new EntityACI[map.size()];
    int i = 0;
    for (final Entry<ResponsibleAgent, Long> e : map.entrySet()) {
      ret[i++] = new EntityACI(e.getKey(), e.getValue() | modBitSet);
    }
    return ret;
  }

  public T create() {
    normalize();
    final ArrayList<EntityACI> acis = new ArrayList<EntityACI>();
    Collections.addAll(acis, toEntityACIArray(this.normal_grants, 0));
    Collections.addAll(acis, toEntityACIArray(this.normal_denials, Long.MIN_VALUE));
    Collections.addAll(acis, toEntityACIArray(this.priority_grants, EntityACL.MIN_PRIORITY_BITSET));
    Collections.addAll(
        acis,
        toEntityACIArray(this.priority_denials, Long.MIN_VALUE | EntityACL.MIN_PRIORITY_BITSET));
    return create(acis);
  }

  private void normalize() {
    for (final Entry<ResponsibleAgent, Long> set : this.priority_denials.entrySet()) {
      if (this.priority_grants.containsKey(set.getKey())) {
        this.priority_grants.put(
            set.getKey(), this.priority_grants.get(set.getKey()) & ~set.getValue());
      }
      if (this.normal_denials.containsKey(set.getKey())) {
        this.normal_denials.put(
            set.getKey(), this.normal_denials.get(set.getKey()) & ~set.getValue());
      }
      if (this.normal_grants.containsKey(set.getKey())) {
        this.normal_grants.put(
            set.getKey(), this.normal_grants.get(set.getKey()) & ~set.getValue());
      }
    }
    for (final Entry<ResponsibleAgent, Long> set : this.priority_grants.entrySet()) {
      if (this.normal_denials.containsKey(set.getKey())) {
        this.normal_denials.put(
            set.getKey(), this.normal_denials.get(set.getKey()) & ~set.getValue());
      }
      if (this.normal_grants.containsKey(set.getKey())) {
        this.normal_grants.put(
            set.getKey(), this.normal_grants.get(set.getKey()) & ~set.getValue());
      }
    }
    for (final Entry<ResponsibleAgent, Long> set : this.normal_denials.entrySet()) {
      if (this.normal_grants.containsKey(set.getKey())) {
        this.normal_grants.put(
            set.getKey(), this.normal_grants.get(set.getKey()) & ~set.getValue());
      }
    }
  }

  public void clear() {
    this.normal_grants.clear();
    this.normal_denials.clear();
    this.priority_grants.clear();
    this.priority_denials.clear();
  }

  protected abstract T create(Collection<EntityACI> acis);
}
