/*
 * ** header v3.0
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2018 Research Group Biomedical Physics,
 * Max-Planck-Institute for Dynamics and Self-Organization Göttingen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */
package caosdb.server.permissions;

import caosdb.server.entity.xml.ToElementable;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.jdom2.Element;

public class EntityPermission extends Permission {

  private static final long serialVersionUID = 1L;
  private static ArrayList<EntityPermission> instances = new ArrayList<EntityPermission>();
  private final int bitNumber;

  public static ToElementable getAllEntityPermissions() {
    final Element entityPermissionsElement = new Element("EntityPermissions");
    for (final EntityPermission e : instances) {
      final Element element = e.toElement();
      element.setAttribute("description", e.getDescription());
      entityPermissionsElement.addContent(element);
    }

    entityPermissionsElement.addContent(EntityACL.GLOBAL_PERMISSIONS.toElement());

    return new ToElementable() {

      @Override
      public void addToElement(final Element ret) {
        ret.addContent(entityPermissionsElement);
      }
    };
  }

  private EntityPermission(final String shortName, final String description, final int bitNumber) {
    super(shortName, description);
    this.bitNumber = bitNumber;
    if (bitNumber > 61) {
      throw new RuntimeException(
          "This bitNumber is too big. This implementation only handles bitNumbers up to 61.");
    }
    if (instances.contains(this)) {
      throw new RuntimeException("This EntityPermission is defined yet.");
    } else {
      instances.add(this);
    }
  }

  public static Set<EntityPermission> getPermissionsPerWildCard(final String s) {
    final Pattern pattern = Pattern.compile(s.replaceAll("\\*", ".*"));

    final HashSet<EntityPermission> ret = new HashSet<EntityPermission>();
    for (final EntityPermission p : instances) {
      final Matcher m = pattern.matcher(p.getShortName());
      if (m.matches()) {
        ret.add(p);
      }
    }
    return ret;
  }

  public static EntityPermission getEntityPermission(final String name) {
    for (final EntityPermission p : instances) {
      if (p.getShortName().equals(name)) {
        return p;
      }
    }
    throw new IllegalArgumentException("Permission is not defined.");
  }

  public static EntityPermission getEntityPermission(final int bitNumber) {
    for (final EntityPermission p : instances) {
      if (p.getBitNumber() == bitNumber) {
        return p;
      }
    }
    throw new IllegalArgumentException("Permission is not defined.");
  }

  public long getBitSet() {
    return (long) Math.pow(2, getBitNumber());
  }

  public int getBitNumber() {
    return this.bitNumber;
  }

  public Element toElement() {
    final Element ret = new Element("Permission");
    ret.setAttribute("name", getShortName());
    return ret;
  }

  public static final EntityPermission RETRIEVE_ENTITY =
      new EntityPermission(
          "RETRIEVE:ENTITY",
          "Permission to retrieve the full entity (name, description, data type, ...) with all parents and properties (unless prohibited by another rule on the property level).",
          4);
  public static final EntityPermission RETRIEVE_ACL =
      new EntityPermission(
          "RETRIEVE:ACL", "Permission to retrieve the full and final ACL of this entity.", 5);
  public static final EntityPermission RETRIEVE_HISTORY =
      new EntityPermission(
          "RETRIEVE:HISTORY", "Permission to retrieve the history of this entity.", 6);
  public static final EntityPermission RETRIEVE_OWNER =
      new EntityPermission(
          "RETRIEVE:OWNER", "Permission to retrieve the owner(s) of this entity.", 9);
  public static final EntityPermission RETRIEVE_FILE =
      new EntityPermission(
          "RETRIEVE:FILE", "Permission to download the file belonging to this entity.", 10);
  public static final EntityPermission DELETE =
      new EntityPermission("DELETE", "Permission to delete an entity.", 1);
  public static final EntityPermission EDIT_ACL =
      new EntityPermission(
          "EDIT:ACL",
          "Permission to change the user-specified part of this entity's ACL. Roles with this Permission are called 'Owners'.",
          0);
  public static final EntityPermission UPDATE_DESCRIPTION =
      new EntityPermission(
          "UPDATE:DESCRIPTION", "Permission to change the value of this entity.", 11);
  public static final EntityPermission UPDATE_VALUE =
      new EntityPermission("UPDATE:VALUE", "Permission to change the value of this entity.", 12);
  public static final EntityPermission UPDATE_ROLE =
      new EntityPermission("UPDATE:ROLE", "Permission to change the role of this entity.", 13);
  public static final EntityPermission UPDATE_REMOVE_PARENT =
      new EntityPermission(
          "UPDATE:PARENT:REMOVE", "Permission  to remove parents from this entity.", 14);
  public static final EntityPermission UPDATE_ADD_PARENT =
      new EntityPermission("UPDATE:PARENT:ADD", "Permission to add a parent to this entity.", 15);
  public static final EntityPermission UPDATE_REMOVE_PROPERTY =
      new EntityPermission(
          "UPDATE:PROPERTY:REMOVE", "Permission to remove properties from this entity.", 16);
  public static final EntityPermission UPDATE_ADD_PROPERTY =
      new EntityPermission(
          "UPDATE:PROPERTY:ADD", "Permission to add a property to this entity.", 17);
  public static final EntityPermission UPDATE_NAME =
      new EntityPermission("UPDATE:NAME", "Permission to change the name of this entity.", 19);
  public static final EntityPermission UPDATE_DATA_TYPE =
      new EntityPermission(
          "UPDATE:DATA_TYPE", "Permission to change the data type of this entity.", 20);
  public static final EntityPermission UPDATE_REMOVE_FILE =
      new EntityPermission(
          "UPDATE:FILE:REMOVE", "Permission to delete the file of this entity.", 21);
  public static final EntityPermission UPDATE_ADD_FILE =
      new EntityPermission("UPDATE:FILE:ADD", "Permission to set a file for this entity.", 22);
  public static final EntityPermission UPDATE_MOVE_FILE =
      new EntityPermission(
          "UPDATE:FILE:MOVE", "Permission to move an existing file to a new location.", 23);
  public static final EntityPermission USE_AS_REFERENCE =
      new EntityPermission(
          "USE:AS_REFERENCE", "Permission to refer to this entity via a reference property.", 24);
  public static final EntityPermission USE_AS_PROPERTY =
      new EntityPermission(
          "USE:AS_PROPERTY", "Permission to implement this entity as a property.", 25);
  public static final EntityPermission USE_AS_PARENT =
      new EntityPermission(
          "USE:AS_PARENT", "Permission to use this entity as a super type for other entities.", 26);
  public static final EntityPermission USE_AS_DATA_TYPE =
      new EntityPermission(
          "USE:AS_DATA_TYPE",
          "Permission to use this entity as a data type for reference properties.",
          27);
  public static final EntityPermission UPDATE_QUERY_TEMPLATE_DEFINITION =
      new EntityPermission(
          "UPDATE:QUERY_TEMPLATE_DEFINITION",
          "Permission to update the query template definition of this QueryTemplate",
          28);
}
