/*
 * ** header v3.0
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2018 Research Group Biomedical Physics,
 * Max-Planck-Institute for Dynamics and Self-Organization Göttingen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */
package caosdb.server.query;

import static java.sql.Types.VARCHAR;

import caosdb.server.query.Query.QueryException;
import java.sql.CallableStatement;
import java.sql.SQLException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.jdom2.Element;

public class StoredAt implements EntityFilterInterface {

  public static final java.util.regex.Pattern PATTERN =
      Pattern.compile("((?:(?:\\\\[\\*\\\\])|[^\\*\\\\])+)?(\\*{2,})?(\\*)?");
  private final String location;
  private boolean pattern_matching = false;

  public StoredAt(final String loc) {
    final String[] dirs = loc.split("/");
    final StringBuffer sb = new StringBuffer();

    // if (dirs[0].length() == 0) {
    // sb.append("/");
    // }
    for (final String dir : dirs) {
      final Matcher m = PATTERN.matcher(dir);

      sb.append("/");
      while (m.find()) {
        // part of a directory name
        if (m.group(1) != null) {
          sb.append(m.group(1).replace("%", "\\%").replace("_", "\\_").replaceAll("\\\\\\*", "*"));
        }

        // **
        if (m.group(2) != null) {
          this.pattern_matching = true;
          sb.append("%");
        }

        // *
        if (m.group(3) != null) {
          this.pattern_matching = true;
          sb.append("%%");
        }
      }
    }

    if (loc.endsWith("/")) {
      sb.append("/");
    }

    this.location =
        sb.toString().replaceFirst("^/", "").replaceFirst("^%{2,}", "%").replaceFirst("^/", "");
  }

  @Override
  public void apply(final QueryInterface query) throws QueryException {
    final long t1 = System.currentTimeMillis();
    try {
      final CallableStatement callSAT = query.getConnection().prepareCall("call applySAT(?,?,?,?)");

      // The wildcards for directories are a bit tricky: * matches
      // anything
      // but the slash. But there is no such option in ordinary MySQL
      // LIKE.
      // Therefore we first include everything which matches an ordinary
      // LIKE
      // ...
      callSAT.setString(1, query.getSourceSet()); // sourceSet
      if (query.getTargetSet() != null) { // targetSet
        callSAT.setString(2, query.getTargetSet());
      } else {
        callSAT.setNull(2, VARCHAR);
      }
      if (this.location == null) { // location
        callSAT.setNull(3, VARCHAR);
      } else {
        callSAT.setString(3, this.location);
      }

      callSAT.setString(4, (this.pattern_matching ? "LIKE" : "="));
      callSAT.execute();

      // ... and then cancel out the wrong guys via regexp
      if (this.pattern_matching && this.location.contains("%%")) {

        // make a MySQL-conform regexp for this.location:
        // 1) escape literal dots (. = \.)
        // 2) use Pattern.quote()
        // 3) replace \Q with ^ (MySQL doesn't know \Q)
        // 4) replace \E with $ (MySQL doesn't know \E)
        // 5) replace non-escaped %% with [^/]*
        // 6) replace non-escaped % with .*
        final String nloc =
            Pattern.quote(this.location.replace(".", "\\."))
                .replaceFirst("..$", "\\$")
                .replaceFirst("^..", "^")
                .replaceAll("(?<!\\\\)%%", "[^/]*")
                .replaceAll("(?<!\\\\)%", ".*");

        if (query.getTargetSet() != null) {
          callSAT.setString(1, query.getTargetSet());
        } else {
          callSAT.setString(1, query.getSourceSet());
        }
        callSAT.setNull(2, VARCHAR);
        callSAT.setString(3, nloc);
        callSAT.setString(4, "RLIKE");
        callSAT.execute();
      }
      callSAT.close();
    } catch (final SQLException e) {
      throw new QueryException(e);
    }
    query.addBenchmark(this.getClass().getSimpleName(), System.currentTimeMillis() - t1);
  }

  @Override
  public Element toElement() {
    final Element ret = new Element("StoredAt");
    ret.setAttribute("location", this.location);
    return ret;
  }

  @Override
  public String toString() {
    return "SAT(" + this.location + ")";
  }
}
